import ctypes
import functools
import threading
from typing import Any, Callable, Union

from subsearch import core
from subsearch.data import __guid__
from subsearch.data.constants import FILE_PATHS
from subsearch.utils import exceptions, io_log, io_toml


def apply_mutex(func: Callable) -> Callable:
    def inner(*args, **kwargs):
        try:
            if not io_toml.load_toml_value(FILE_PATHS.config, "misc.single_instance"):
                return func()
        except FileNotFoundError:
            pass
        except KeyError:
            pass
        except TypeError:
            pass
        kernel32 = ctypes.WinDLL("kernel32")
        mutex = kernel32.CreateMutexW(None, False, __guid__)
        last_error = kernel32.GetLastError()

        if last_error == 183:
            raise exceptions.MultipleInstancesError(__guid__)
        try:
            kernel32.WaitForSingleObject(mutex, -1)
            return func(*args, **kwargs)
        finally:
            kernel32.ReleaseMutex(mutex)

    return inner


def check_option_disabled(func):
    def wrapper(self, event, *args, **kwargs):
        btn = event.widget
        if btn.instate(["disabled"]):
            return None
        return func(self, event, *args, **kwargs)

    return wrapper


def singleton(cls):
    previous_instances: dict[Callable, Any] = {}

    @functools.wraps(cls)
    def wrapper(*args, **kwargs):
        if cls in previous_instances and previous_instances.get(cls, None).get("args") == (args, kwargs):
            return previous_instances[cls].get("instance")
        else:
            previous_instances[cls] = {"args": (args, kwargs), "instance": cls(*args, **kwargs)}
            return previous_instances[cls].get("instance")

    return wrapper


def call_conditions(func):
    def wrapper(*args, **kwargs):
        function = f"{func.__name__}"
        if not CallCondition.conditions(function=function, *args, **kwargs):
            module_fn = f"{func.__module__}.{func.__name__}"
            io_log.stdout(f"Call conditions for '{module_fn}', not met", level="debug", print_allowed=False)
            return None
        return func(*args, **kwargs)

    return wrapper


def thread_safe_log(func):
    lock = threading.Lock()

    def wrapper_log(cls, *args, **kwargs):
        with lock:
            return func(cls, *args, **kwargs)

    return wrapper_log


class CallCondition:
    @staticmethod
    def language_compatibility(provider: str):
        language = io_toml.load_toml_value(FILE_PATHS.config, "subtitle_filters.language")
        incompatibility: list[str] = io_toml.load_toml_value(FILE_PATHS.language_data, f"{language}.incompatibility")
        if provider in incompatibility:
            return False
        return True

    @staticmethod
    def all_conditions_met(conditions: list[bool]) -> bool:
        if all(condition for condition in conditions):
            return True
        return False

    @staticmethod
    def conditions(cls: Union["core.SubsearchCore", "core.Initializer"], *args, **kwargs) -> bool:
        if not cls.file_exist:
            return False
        function = kwargs["function"]
        conditions = {
            "opensubtitles": [
                CallCondition.language_compatibility("opensubtitles"),
                cls.app_config.providers["opensubtitles_hash"] or cls.app_config.providers["opensubtitles_site"],
            ],
            "subscene": [
                CallCondition.language_compatibility("subscene"),
                cls.app_config.providers["subscene_site"],
            ],
            "yifysubtitles": [
                not cls.app_config.only_foreign_parts,
                CallCondition.language_compatibility("yifysubtitles"),
                not cls.release_data.tvseries,
                not cls.provider_urls.yifysubtitles == "",
                cls.app_config.providers["yifysubtitles_site"],
            ],
            "download_files": [len(cls.accepted_subtitles) >= 1],
            "manual_download": [
                len(cls.accepted_subtitles) == 0,
                len(cls.rejected_subtitles) >= 1,
                cls.app_config.manual_download_on_fail,
            ],
            "extract_files": [len(cls.accepted_subtitles) >= 1],
            "subtitle_post_processing": [],
            "subtitle_rename": [cls.app_config.subtitle_post_processing["rename"], cls.subtitles_found > 1],
            "subtitle_move_best": [
                cls.app_config.subtitle_post_processing["move_best"],
                cls.subtitles_found > 1,
                not cls.app_config.subtitle_post_processing["move_all"],
            ],
            "subtitle_move_all": [cls.app_config.subtitle_post_processing["move_all"], cls.subtitles_found > 1],
            "summary_notification": [cls.app_config.summary_notification],
            "clean_up": [],
        }
        return CallCondition.all_conditions_met(conditions[function])


enable_system_tray: bool


def system_tray_conditions(func):
    def wrapper(*args, **kwargs):
        if enable_system_tray:
            return func(*args, **kwargs)

    return wrapper



from selectolax.parser import Node

from subsearch.data.data_classes import Subtitle
from subsearch.providers import core_provider


class YifySubtitlesScraper(core_provider.ProviderHelper):
    def __init__(self, **kwargs) -> None:
        core_provider.ProviderHelper.__init__(self, **kwargs)

    def skip_item(self, item: Node, hi_sub: bool, regular_sub: bool, current_language: str) -> bool:
        subtitle_language = item.css_first("span.sub-lang").child.text_content
        hearing_impaired = item.css_matches("span.hi-subtitle")
        if subtitle_language.lower() != current_language.lower():
            return True
        if (hi_sub and regular_sub) or (hi_sub is False and regular_sub is False):
            pass
        elif hi_sub is False and hearing_impaired:
            return True
        elif regular_sub is True and hearing_impaired:
            return True
        return False

    def get_subtitle(self, url: str, current_language: str, hi_sub: bool, non_hi_sub: bool) -> dict[str, str]:
        subtitles: dict[str, str] = {}
        tree = core_provider.get_html_parser(url)
        product = tree.select("tr")
        for item in product.matches[1:]:
            if self.skip_item(item, hi_sub, non_hi_sub, current_language):
                continue
            node = item.css_first("a")
            titles = node.text().strip().split("subtitle ")[-1].split("\n")
            _href = node.attributes["href"].split("/")
            href = _href[-1]
            for title in titles:
                subtitles[title] = f"https://yifysubtitles.org/subtitle/{href}.zip"
        return subtitles


class YifiSubtitles(YifySubtitlesScraper):
    def __init__(self, **kwargs):
        YifySubtitlesScraper.__init__(self, **kwargs)
        self.provider_name = self.__class__.__name__.lower()

    def start_search(self) -> list[Subtitle] | None:
        subtitle_data = self.get_subtitle(self.url_yifysubtitles, self.current_language, self.hi_sub, self.non_hi_sub)

        if not subtitle_data:
            return []

        self._process_subtitle_data(self.provider_name, subtitle_data)

    @property
    def accepted_subtitles(self) -> list[Subtitle]:
        return self._accepted_subtitles

    @property
    def rejected_subtitles(self) -> list[Subtitle]:
        return self._rejected_subtitles



import re
from typing import Any

from subsearch.data.data_classes import Subtitle
from subsearch.providers import core_provider
from subsearch.utils import io_log


class OpenSubtitlesScraper(core_provider.ProviderHelper):
    def __init__(self, **kwargs) -> None:
        core_provider.ProviderHelper.__init__(self, **kwargs)

    def opensubtitles_down(self, tree: Any):
        is_offline = tree.css_matches("pre")
        if is_offline is False:
            return False
        offline_text = tree.css_first("pre").text()
        if offline_text.startswith("Site will be online soon"):
            io_log.stdout(f"opensubtitles is down: {offline_text}", level="error")
            return True
        return False

    def get_subtitles(self, url: str):
        subtitles: dict[str, str] = {}
        tree = core_provider.get_html_parser(url)
        items = tree.css("item")
        if self.opensubtitles_down(tree):
            return subtitles
        for item in items:
            dl_url = item.css_first("enclosure").attributes["url"]
            released_as = item.css_first("description").child.text_content.strip()
            release_name = re.findall("^.*?: (.*?);", released_as)[0]  # https://regex101.com/r/LWAmJK/1
            subtitles[release_name] = dl_url
        return subtitles

    def with_hash(self, url: str, release: str) -> dict[str, str]:
        subtitles: dict[str, str] = {}
        tree = core_provider.get_html_parser(url)
        if self.opensubtitles_down(tree):
            return subtitles
        try:
            sub_id = tree.css_first("#bt-dwl-bt").attributes["data-product-id"]
        except AttributeError:
            return subtitles
        subtitles[release] = f"https://dl.opensubtitles.org/en/download/sub/{sub_id}"
        return subtitles


class OpenSubtitles(OpenSubtitlesScraper):
    def __init__(self, **kwargs):
        OpenSubtitlesScraper.__init__(self, **kwargs)
        self.provider_name = self.__class__.__name__.lower()

    def start_search(self, flag: str) -> None:
        flags = {"hash": self.search_site, "site": self.search_hash}
        flags[flag]()

    def search_hash(self) -> None:
        subtitle_data = self.with_hash(self.url_opensubtitles_hash, self.release)

        if not subtitle_data:
            return None

        io_log.stdout_match(
            provider=self.provider_name,
            subtitle_name=self.release,
            result=100,
            threshold=self.app_config.accept_threshold,
        )
        self._process_subtitle_data(self.provider_name, subtitle_data)

    def search_site(self) -> None:
        subtitle_data = self.get_subtitles(self.url_opensubtitles)

        if not subtitle_data:
            return None

        self._process_subtitle_data(self.provider_name, subtitle_data)

    @property
    def accepted_subtitles(self) -> list[Subtitle]:
        return self._accepted_subtitles

    @property
    def rejected_subtitles(self) -> list[Subtitle]:
        return self._rejected_subtitles



from dataclasses import dataclass
from enum import Enum, auto
from typing import Generator, Union

from subsearch.utils import io_log


@dataclass(order=True, frozen=True)
class DataState(Enum):
    UNKNOWN = auto()
    INITIALIZING_APP_PATHS = auto()
    INITIALIZING_FILE_DATA = auto()
    INITIALIZING_DEVICE_INFO = auto()


@dataclass(order=True, frozen=True)
class CoreState(Enum):
    UNKNOWN = auto()
    INITIALIZE = auto()
    INITIALIZED = auto()
    CALL_OPENSUBTITLES = auto()
    CALL_SUBSCENE = auto()
    CALL_YIFYSUBTITLES = auto()
    DOWNLOAD_FILES = auto()
    MANUAL_DOWNLOAD = auto()
    AUTOLOAD_RENAME = auto()
    AUTOLOAD_MOVE = auto()
    AUTOLOAD_MOVE_ALL = auto()
    EXTRACT_FILES = auto()
    SUMMARY_TOAST = auto()
    CLEAN_UP = auto()
    EXIT = auto()

    NO_FILE = auto()
    GUI = auto()


@dataclass(order=True, frozen=True)
class SubsceneState(Enum):
    UNKNOWN = auto()
    WORKING = auto()
    FINNISHED = auto()


@dataclass(order=True, frozen=True)
class OpenSubtitlesState(Enum):
    UNKNOWN = auto()
    WORKING = auto()
    FINNISHED = auto()


@dataclass(order=True, frozen=True)
class YifySubtitlesState(Enum):
    UNKNOWN = auto()
    WORKING = auto()
    FINNISHED = auto()


class Singleton(type):
    _instances: dict[type, type] = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super().__call__(*args, **kwargs)
        return cls._instances[cls]


class StateManager(metaclass=Singleton):
    current_state: Enum
    enum_class = type[Union[DataState, CoreState, SubsceneState, OpenSubtitlesState, YifySubtitlesState]]
    state_history: list[Enum] = []

    def __init__(self, _enum_class: enum_class) -> None:
        self.current_state = self.set_state(_enum_class.UNKNOWN)

    def _yield_state(self) -> Generator:
        current_state = self.current_state
        state_names = [state for state in self.current_state.__class__]
        current_index = state_names.index(current_state)

        next_index = current_index + 1
        if next_index >= len(state_names):
            next_index = 0

        self.set_state(state_names[next_index])
        yield self.current_state

    def set_state(self, state: Enum):
        self.current_state = state
        self.state_history.append(self.current_state)
        io_log.stdout(self.current_state, level="debug", print_allowed=False)

    def get_state(self) -> Enum:
        return self.current_state

    def get_state_history(self) -> list[Enum]:
        return self.state_history

    def update_state(self) -> None:
        next(self._yield_state())


class DataStateManager(StateManager):
    def __init__(self) -> None:
        self.state = DataState
        super().__init__(self.state)


class CoreStateManager(StateManager):
    def __init__(self) -> None:
        self.state = CoreState
        super().__init__(self.state)


class SubsceneStateManager(StateManager):
    def __init__(self) -> None:
        self.state = SubsceneState
        super().__init__(self.state)


class OpenSubtitlesStateManager(StateManager):
    def __init__(self) -> None:
        self.state = OpenSubtitlesState
        super().__init__(self.state)


class YifySubtitlesStateManager(StateManager):
    def __init__(self) -> None:
        self.state = YifySubtitlesState
        super().__init__(self.state)


def initialize_manager(instance):
    manager = instance()
    manager.set_state(manager.state.WORKING)



from pathlib import Path

import cloudscraper
from selectolax.parser import HTMLParser

from subsearch.data.constants import FILE_PATHS, VIDEO_FILE
from subsearch.data.data_classes import (
    AppConfig,
    LanguageData,
    ProviderUrls,
    ReleaseData,
    SubsceneCookie,
    Subtitle,
)
from subsearch.utils import io_log, io_toml, string_parser


class CustomSubsceneHeader:
    def __init__(self, app_config: AppConfig) -> None:
        self.app_config = app_config

    def _get_hearing_impaired_int(self) -> int:
        if self.app_config.hearing_impaired and self.app_config.non_hearing_impaired:
            value = 0
        if self.app_config.hearing_impaired and not self.app_config.non_hearing_impaired:
            value = 1
        if self.app_config.non_hearing_impaired and not self.app_config.hearing_impaired:
            value = 2
        return value

    def _get_cookie(self) -> SubsceneCookie:
        languages = io_toml.load_toml_data(FILE_PATHS.language_data)
        subscene_id = languages[self.app_config.language]["subscene_id"]
        subscene_cookie = SubsceneCookie(
            dark_theme=False,
            sort_subtitle_by_date="false",
            language_filter=subscene_id,
            hearing_impaired=self._get_hearing_impaired_int(),
            foreigen_only=self.app_config.only_foreign_parts,
        )
        return subscene_cookie

    def set_cookie_values(self) -> str:
        data = self._get_cookie()
        dark_theme = f"DarkTheme={data.dark_theme}"
        sort_subtitle_by_date = f"SortSubtitlesByDate={data.sort_subtitle_by_date}"
        language_filter = f"LanguageFilter={data.language_filter}"
        hearing_impaired = f"HearingImpaired={data.hearing_impaired}"
        foreign_only = f"ForeignOnly={data.foreigen_only}"
        return f"{dark_theme}; {sort_subtitle_by_date}; {language_filter}; {hearing_impaired}; {foreign_only}"

    def create_header(self) -> dict[str, str]:
        cookie_values = self.set_cookie_values()
        return {"Cookie": cookie_values}


class SearchArguments:
    def __init__(self, **kwargs):
        release_data: ReleaseData = kwargs["release_data"]
        app_config: AppConfig = kwargs["app_config"]
        provider_urls: ProviderUrls = kwargs["provider_urls"]
        language_data: LanguageData = kwargs["language_data"]

        self.app_config = app_config

        # file parameters
        self.title = release_data.title
        self.year = release_data.year
        self.season = release_data.season
        self.season_ordinal = release_data.season_ordinal
        self.episode = release_data.episode
        self.episode_ordinal = release_data.episode_ordinal
        self.tvseries = release_data.tvseries
        self.release = release_data.release
        self.group = release_data.group

        # user parameters
        self.current_language = app_config.language
        self.hi_sub = app_config.hearing_impaired
        self.non_hi_sub = app_config.non_hearing_impaired
        self.percentage_threashold = app_config.accept_threshold
        self.manual_download_on_fail = app_config.manual_download_on_fail
        # provider url data
        self.url_subscene = provider_urls.subscene
        self.url_opensubtitles = provider_urls.opensubtitles
        self.url_opensubtitles_hash = provider_urls.opensubtitles_hash
        self.url_yifysubtitles = provider_urls.yifysubtitles

        self.language_data = language_data
        self.filehash = VIDEO_FILE.file_hash


class ProviderHelper(SearchArguments):
    def __init__(self, **kwargs):
        SearchArguments.__init__(self, **kwargs)
        self._accepted_subtitles: list[Subtitle] = []
        self._rejected_subtitles: list[Subtitle] = []

    def _process_subtitle_data(self, provider_name: str, subtitle_data: dict[str, str]):
        if not subtitle_data:
            return None

        for subtitle_name, subtitle_url in subtitle_data.items():
            pct_result = string_parser.calculate_match(subtitle_name, self.release)
            io_log.stdout_match(
                provider=provider_name,
                subtitle_name=subtitle_name,
                result=pct_result,
                threshold=self.app_config.accept_threshold,
            )
            if is_threshold_met(self, pct_result):
                if provider_name == "subscene":
                    subtitle_url = self.subscene_get_download_url(subtitle_url)
                subtitle = Subtitle(pct_result, provider_name, subtitle_name.lower(), subtitle_url)
                self._accepted_subtitles.append(subtitle)
            else:
                # * 'get_download_url' on subtitle_url from subscene
                # Would take to long to scrape all subtitle download urls
                subtitle = Subtitle(pct_result, provider_name, subtitle_name.lower(), subtitle_url)
                self._rejected_subtitles.append(subtitle)

    @staticmethod
    def subscene_get_download_url(url: str) -> str:
        tree = get_html_parser(url)
        href = tree.css_first("#downloadButton").attributes["href"]
        download_url = f"https://subscene.com/{href}"
        return download_url


def is_threshold_met(cls: "SearchArguments", pct_result: int) -> bool:
    if pct_result >= cls.percentage_threashold:
        return True
    return False


def get_cloudscraper():
    return cloudscraper.create_scraper(browser={"browser": "chrome", "platform": "android", "desktop": False})


def get_html_parser(url: str, header_=None):
    scraper = get_cloudscraper()
    if header_ is None:
        response = scraper.get(url)
    else:
        response = scraper.get(url, headers=header_)
    return HTMLParser(response.text)



import sys
import time
from pathlib import Path
from threading import Thread

from subsearch import core
from subsearch.data import __guid__
from subsearch.data.constants import FILE_PATHS
from subsearch.utils import decorators, io_log, io_toml

PREF_COUNTER = time.perf_counter()
PACKAGEPATH = Path(__file__).resolve().parent.as_posix()
HOMEPATH = Path(PACKAGEPATH).parent.as_posix()
sys.path.append(HOMEPATH)
sys.path.append(PACKAGEPATH)


class Subsearch:
    def __init__(self) -> None:
        self.subsearch_core = core.SubsearchCore(PREF_COUNTER)

    def thread_executor(self, *args) -> None:
        provider_threads = {}
        for provider in args:
            provider_threads[provider] = Thread(target=provider)

        for thread in provider_threads.values():
            thread.start()

        for thread in provider_threads.values():
            thread.join()

    def search_for_subtitles(self) -> None:
        providers = [self.subsearch_core.subscene, self.subsearch_core.opensubtitles, self.subsearch_core.yifysubtitles]

        if io_toml.load_toml_value(FILE_PATHS.config, "misc.multithreading"):
            self.thread_executor(*providers)

        else:
            for provider in providers:
                provider()

    def provider_opensubtitles(self) -> None:
        self.subsearch_core.opensubtitles()

    def provider_subscene(self) -> None:
        self.subsearch_core.subscene()

    def provider_yifysubtitles(self) -> None:
        self.subsearch_core.yifysubtitles()

    def process_files(self) -> None:
        self.subsearch_core.download_files()
        self.subsearch_core.manual_download()
        self.subsearch_core.extract_files()
        self.subsearch_core.subtitle_post_processing()
        self.subsearch_core.clean_up()

    def on_exit(self) -> None:
        self.subsearch_core.core_on_exit()


@decorators.apply_mutex
def main() -> None:
    app = Subsearch()
    app.search_for_subtitles()
    app.process_files()
    app.on_exit()


def custom_excepthook(exctype, value, traceback):
    io_log.Logger().debug_logger.error(value, exc_info=(exctype, value, traceback))
    sys.__excepthook__(exctype, value, traceback)


if __name__ == "__main__":
    sys.excepthook = custom_excepthook
    main()



import tkinter as tk
import webbrowser
from tkinter import BooleanVar, ttk
from typing import Any

from subsearch.data import __version__
from subsearch.data.constants import DEVICE_INFO, FILE_PATHS
from subsearch.gui.resources import config as cfg
from subsearch.utils import decorators, io_toml, io_winreg, update


class FileExtensions(ttk.Labelframe):
    instances: list["FileExtensions"] = []

    def __init__(self, parent) -> None:
        ttk.Labelframe.__init__(self, parent)
        self.configure(text="File Exstensions", padding=10)
        FileExtensions.instances.append(self)
        self.data = io_toml.load_toml_data(FILE_PATHS.config)
        self.file_extensions = io_toml.load_toml_value(FILE_PATHS.config, "file_extensions")

        self.checkbuttons: dict[ttk.Checkbutton, tuple[str, BooleanVar]] = {}

        frame = None
        for enum, (key, value) in enumerate(self.file_extensions.items()):
            if enum % 4 == 0:
                frame = ttk.Frame(self)
                frame.pack(side=tk.LEFT, anchor="n")

            boolean = tk.BooleanVar()
            boolean.set(value)
            check_btn = ttk.Checkbutton(frame, text=key, onvalue=True, offvalue=False, variable=boolean)
            if not io_toml.load_toml_value(FILE_PATHS.config, "gui.context_menu"):
                check_btn.state(["disabled"])

            check_btn.pack(padx=4, pady=4, ipadx=10)
            self.checkbuttons[check_btn] = key, boolean
            check_btn.bind("<Enter>", self.enter_button)
        setattr(FileExtensions, "instance_cbtn", self.checkbuttons)

    def enter_button(self, event) -> None:
        btn = event.widget
        btn.bind("<ButtonPress-1>", self.press_button)

    def press_button(self, event) -> None:
        btn = event.widget
        btn.bind("<ButtonRelease-1>", self.toggle_ext)

    @decorators.check_option_disabled
    def toggle_ext(self, event) -> None:
        btn = event.widget
        key = self.checkbuttons[btn][0]
        value = self.checkbuttons[btn][1]
        if value.get():
            self.data["file_extensions"][key] = False
        elif not value.get():
            self.data["file_extensions"][key] = True
        io_toml.dump_toml_data(FILE_PATHS.config, self.data)

        self.update_registry()

    def update_registry(self) -> None:
        io_winreg.write_all_valuex()

    def update_state(self):
        for key in self.checkbuttons.keys():
            if io_toml.load_toml_value(FILE_PATHS.config, "gui.context_menu"):
                key.state(["!disabled"])
            else:
                key.state(["disabled"])


class SubsearchOption(ttk.Labelframe):
    def __init__(self, parent) -> None:
        ttk.Labelframe.__init__(self, parent)
        self.configure(text="Subsearch Options", padding=10)
        self.data = io_toml.load_toml_data(FILE_PATHS.config)
        self.subsearch_options: dict[str, Any] = {
            "gui.context_menu": "Context menu",
            "gui.context_menu_icon": "Context menu icon",
            "gui.system_tray": "System tray icon",
            "gui.summary_notification": "Notification when done",
            "gui.show_terminal": "Terminal while searching",
            "misc.manual_download_on_fail": "Manual download on fail",
            "misc.multithreading": "Multithreading",
            "misc.single_instance": "Single instance",
        }
        for name, description in self.subsearch_options.items():
            self.subsearch_options[name] = [
                io_toml.load_toml_value(FILE_PATHS.config, name),
                description,
            ]

        self.checkbuttons: dict[ttk.Checkbutton, tuple[str, BooleanVar]] = {}
        frame = None

        for enum, (key, value) in enumerate(self.subsearch_options.items()):
            if enum % 3 == 0:
                frame = ttk.Frame(self)
                frame.pack(side=tk.LEFT, anchor="n")

            boolean = tk.BooleanVar()
            boolean.set(value[0])
            check_btn = ttk.Checkbutton(frame, text=value[1], onvalue=True, offvalue=False, variable=boolean)
            context_menu = io_toml.load_toml_value(FILE_PATHS.config, "gui.context_menu")
            system_tray = io_toml.load_toml_value(FILE_PATHS.config, "gui.system_tray")
            if key == "gui.context_menu_icon" and not context_menu:
                check_btn.state(["disabled"])
            if key == "gui.show_terminal" and DEVICE_INFO.mode == "executable":
                check_btn.state(["disabled"])
            if key == "gui.summary_notification" and not system_tray:
                check_btn.state(["disabled"])
            check_btn.pack(padx=4, pady=4, ipadx=40)

            self.checkbuttons[check_btn] = key, boolean
            check_btn.bind("<Enter>", self.enter_button)

    def enter_button(self, event) -> None:
        btn = event.widget
        btn.bind("<ButtonPress-1>", self.press_button)

    def press_button(self, event) -> None:
        btn = event.widget
        btn.bind("<ButtonRelease-1>", self.toggle_btn)

    @decorators.check_option_disabled
    def toggle_btn(self, event) -> None:
        btn = event.widget
        key = self.checkbuttons[btn][0]
        value = self.checkbuttons[btn][1]
        if value.get():
            io_toml.update_toml_key(FILE_PATHS.config, key, False)
        elif not value.get():
            io_toml.update_toml_key(FILE_PATHS.config, key, True)
        self.update_registry(btn)
        keys = [("gui.context_menu", "gui.context_menu_icon"), ("gui.system_tray", "gui.summary_notification")]
        for key_pair in keys:
            self.disable_check_btn_children(btn, value, key_pair)

    def disable_check_btn_children(self, btn: Any, value: BooleanVar, key_pair: tuple[str, str]):
        # FileExtension checkboxes
        parent_key = key_pair[0]
        child_key = key_pair[1]
        if btn["text"] != self.subsearch_options[parent_key][1]:
            return None
        [instance.update_state() for instance in FileExtensions.instances]

        # SubsearchOptions checkboxes
        for check_button, tuple_value in self.checkbuttons.items():
            key = tuple_value[0]
            if key != child_key:
                continue
            elif value.get():
                check_button.state(["disabled"])
            elif not value.get():
                check_button.state(["!disabled"])

    def update_registry(self, btn) -> None:
        if btn["text"] != self.subsearch_options["gui.context_menu"][1]:
            return None
        menu = io_toml.load_toml_value(FILE_PATHS.config, "gui.context_menu")
        if menu:
            io_winreg.add_context_menu()
        else:
            io_winreg.del_context_menu()


class CheckForUpdates(ttk.Labelframe):
    def __init__(self, parent) -> None:
        ttk.Labelframe.__init__(self, parent)
        self.configure(text="Update Subsearch", padding=10, style="TLabelframePlain")
        frame_left = tk.Frame(self, bg=cfg.color.default_bg)
        frame_right = tk.Frame(self, bg=cfg.color.default_bg)

        self.var_current = tk.StringVar(self, f"Current version:\t\t{__version__}")
        self.var_latest = tk.StringVar(self, "Latest version: \t\t")
        self.var_misc = tk.StringVar()

        frame_current = tk.Label(frame_left, textvariable=self.var_current, justify="left")
        frame_latest = tk.Label(frame_left, textvariable=self.var_latest, justify="left")
        self.frame_misc = tk.Label(frame_left, textvariable=self.var_misc, justify="left")
        frame_current.pack(anchor="w")
        frame_latest.pack(anchor="w")
        self.frame_misc.pack(pady=16, anchor="w")

        self.btn_search = ttk.Button(
            frame_right,
            text="Search",
            width=20,
        )
        self.btn_visit_release_page = ttk.Button(
            frame_right,
            text="Open in webbrowser",
            width=20,
        )
        self.btn_search.pack(padx=2, pady=2)
        self.btn_visit_release_page.pack(padx=2, pady=2)
        self.btn_visit_release_page.state(["disabled"])
        self.btn_search.bind("<Enter>", self.enter_button)

        self.pack(anchor="center", fill="x")
        width = self.winfo_reqwidth()
        half_width = round(width // 2)
        frame_left.pack(side=tk.LEFT, anchor="n", expand=True, fill="x", ipadx=half_width)
        frame_right.pack(side=tk.LEFT, anchor="n", expand=True, fill="x", ipadx=half_width)

    def enter_button(self, event) -> None:
        btn = event.widget
        if btn == self.btn_search:
            btn.bind("<ButtonRelease-1>", self.search_button)

    def search_button(self, event) -> None:
        new_repo_avail, repo_is_prerelease = update.is_new_version_avail()
        latest_version = update.get_latest_version()
        if new_repo_avail:
            self.var_latest.set(f"Latest version: \t\t{latest_version}")
            self.btn_search.state(["disabled"])
            self.btn_visit_release_page.state(["!disabled"])
            self.btn_visit_release_page.bind("<ButtonRelease-1>", self.visit_repository_button)
            if repo_is_prerelease:
                self.var_misc.set(f"Pre-release")
                self.frame_misc.configure(fg=cfg.color.orange)
            else:
                self.var_misc.set(f"Latest")
                self.frame_misc.configure(fg=cfg.color.green)

        if not new_repo_avail:
            self.var_latest.set(f"Latest version: \t\t{latest_version}")
            self.var_misc.set(f"Latest version already installed")

    def visit_repository_button(self, event) -> None:
        webbrowser.open(f"https://github.com/vagabondHustler/subsearch/releases")



import shutil
import struct
import zipfile
from io import BufferedReader
from pathlib import Path
from typing import Any

from subsearch.data.constants import VIDEO_FILE
from subsearch.data.data_classes import Subtitle
from subsearch.providers.core_provider import get_cloudscraper
from subsearch.utils import io_log, string_parser


def create_path_from_string(string: str, path_resolution: str) -> Path:
    if path_resolution == "relative":
        if string == ".":
            path = VIDEO_FILE.file_directory
        elif string.startswith(".\\") and len(string) > 2:
            path = VIDEO_FILE.file_directory.joinpath(string[2:])
        elif string == "..":
            path = VIDEO_FILE.file_directory.parent
        elif string.startswith("..\\") and len(string) > 3:
            path = VIDEO_FILE.file_directory.parent.joinpath(string[3:])
    elif path_resolution == "absolute":
        path = Path(string)

    path.mkdir(parents=True, exist_ok=True)
    return path


def download_subtitle(subtitle: Subtitle, index_position: int, index_size: int):
    io_log.stdout(f"{subtitle.provider}: {index_position}/{index_size}: {subtitle.release_name}")
    scraper = get_cloudscraper()
    r = scraper.get(subtitle.download_url, stream=True)
    file_name = f"{subtitle.provider}_{subtitle.release_name}_{index_position}.zip"
    download_path = VIDEO_FILE.tmp_dir / file_name
    with open(download_path, "wb") as fd:
        for chunk in r.iter_content(chunk_size=1024):
            fd.write(chunk)


def extract_files_in_dir(src: Path, dst: Path, extension: str = ".zip") -> None:
    for file in src.glob(f"*{extension}"):
        filename = src / file
        io_log.stdout_path_action(action_type="extract", src=file, dst=dst)
        zip_ref = zipfile.ZipFile(filename)
        zip_ref.extractall(dst)
        zip_ref.close()


def autoload_rename(release_name: str, extension: str = ".srt") -> Path:
    best_match = (0, Path("."))
    for file in VIDEO_FILE.subs_dir.glob(f"*{extension}"):
        value = string_parser.calculate_match(file.name, release_name)
        if value >= best_match[0]:
            best_match = value, file

    old_file_path = best_match[1]
    new_file_path = old_file_path.with_name(f"{release_name}{extension}")
    io_log.stdout_path_action(action_type="rename", src=old_file_path, dst=new_file_path)
    old_file_path.rename(new_file_path)
    return new_file_path


def move_all(src: Path, dst: Path, extension: str = ".srt"):
    for file in src.glob(f"*{extension}"):
        move_and_replace(file.absolute(), dst)


def move_and_replace(source_file: Path, destination_directory: Path) -> None:
    source_file.replace(destination_directory / source_file.name)


def del_file_type(cwd: Path, extension: str) -> None:
    for file in Path(cwd).glob(f"*{extension}"):
        io_log.stdout_path_action(action_type="remove", src=file)
        file_path = Path(cwd) / file
        file_path.unlink()


def del_directory(directory: Path) -> None:
    io_log.stdout_path_action(action_type="remove", src=directory)
    shutil.rmtree(directory)


def directory_is_empty(directory: Path) -> bool:
    if not any(directory.iterdir()):
        return True
    return False


def del_directory_content(directory: Path):
    for item in directory.iterdir():
        io_log.stdout_path_action(action_type="remove", src=item)
        if item.is_file():
            item.unlink()
        if item.is_dir():
            shutil.rmtree(item)


def create_directory(path: Path):
    path.mkdir(exist_ok=True)


def get_file_hash(file_path: Path) -> str:
    if not file_path:
        return ""

    hash_algorithm = MPCHashAlgorithm(file_path)
    return hash_algorithm.get_hash()


class MPCHashAlgorithm:
    def __init__(self, file_path: Path, **kwargs) -> None:
        file_size = file_path.stat().st_size
        self.file_path = file_path
        self.chunk_size = kwargs.get("chunk_size", 65536)
        self.file_size, self.hash_chunk = file_size, file_size
        self.hash = self.mpc_hash_algorithm()

    def mpc_hash_algorithm(self) -> str:
        if not self.valid_file_size():
            return ""

        with open(self.file_path, "rb") as file:
            self.byte_size = struct.calcsize("<q")
            self.update_hash_chunk(file)
            file.seek(max(0, self.file_size - self.chunk_size), 0)
            self.update_hash_chunk(file)
        return "%016x" % self.hash_chunk

    def valid_file_size(self) -> bool:
        if self.file_size < self.chunk_size * 2:
            io_log.stdout(f"Invalid file size, {self.file_size} bytes", level="warning")
            return False
        return True

    def update_hash_chunk(self, file: BufferedReader) -> None:
        for _ in range(self.chunk_size // self.byte_size):
            buffer = file.read(self.byte_size)
            (chunk_value,) = struct.unpack("<q", buffer)
            self.hash_chunk += chunk_value
            self.hash_chunk = self.hash_chunk & 0xFFFFFFFFFFFFFFFF

    def get_hash(self) -> str:
        return self.hash



import tkinter as tk
from tkinter import ttk

from subsearch.data.constants import VIDEO_FILE
from subsearch.data.data_classes import Subtitle
from subsearch.gui.resources import config as cfg
from subsearch.providers import core_provider
from subsearch.utils import io_file_system, io_log, string_parser


class DownloadManager(ttk.LabelFrame):
    downloaded_subtitle: list[Subtitle] = []

    def __init__(self, parent, subtitles: list[Subtitle]) -> None:
        ttk.Labelframe.__init__(self, parent)
        self.configure(text="Available subtitles", padding=10)
        if subtitles:
            subtitles.sort(key=lambda x: x.pct_result, reverse=True)
        frame_left = tk.Frame(self)
        frame_sep = tk.Frame(self, width=5)
        frame_right = tk.Frame(self)
        frame_left.pack(side=tk.LEFT, expand=True, fill="both")
        frame_sep.pack(side=tk.LEFT, fill="y")
        frame_right.pack(side=tk.LEFT, fill="y")
        self.failed_subtitle_downloads: list[Subtitle] = []
        self.download_number = 1
        self.download_index_size = len(subtitles)
        self.subtitles = subtitles
        self.scrollbar = ttk.Scrollbar(frame_right, orient="vertical", style="Vertical.TScrollbar")
        self.sub_listbox = tk.Listbox(
            frame_left,
            height=36,
            bg=cfg.color.default_bg,
            fg=cfg.color.light_grey,
            font=cfg.font.cas8b,
            bd=0,
            border=0,
            borderwidth=0,
            highlightthickness=0,
            activestyle="none",
            yscrollcommand=self.scrollbar.set,
        )
        self.sub_listbox.pack(expand=True, fill="both")
        if self.subtitles is not None:
            self.fill_listbox()
        self.scrollbar.pack(expand=True, fill="y")
        self.scrollbar.config(command=self.sub_listbox.yview)

    def fill_listbox(self) -> None:
        self.listbox_index: dict[int, Subtitle] = {}
        for enum, subtitle in enumerate(self.subtitles):
            self.sub_listbox.insert(tk.END, f"{subtitle.pct_result}% {subtitle.release_name}\n")
            self.sub_listbox.bind("<ButtonPress-1>", self.mouse_b1_press)
            self.listbox_index[enum] = subtitle

    def mouse_b1_press(self, event) -> None:
        self.sub_listbox.unbind("<ButtonPress-1>")
        self.sub_listbox.bind("<<ListboxSelect>>", self.select_subtitle)

    def select_subtitle(self, event) -> None:
        self.sub_listbox.unbind("<<ListboxSelect>>")
        _selection = self.sub_listbox.curselection()
        selection = _selection[0]
        subtitle = self.listbox_index[selection]
        if subtitle in (self.downloaded_subtitle or self.failed_subtitle_downloads):
            print("i'm here")
            self.sub_listbox.bind("<ButtonPress-1>", self.mouse_b1_press)
            return
        if subtitle.provider == "subscene":
            subtitle.download_url = core_provider.ProviderHelper.subscene_get_download_url(subtitle.download_url)
        self.update_text(selection, "⊙", subtitle, cfg.color.orange)
        self.sub_listbox.bind("<ButtonRelease-1>", lambda event: self.download(event, subtitle, selection))

    def download(self, event, subtitle: Subtitle, selection: int) -> None:
        self.sub_listbox.unbind("<ButtonRelease-1>")
        try:
            if string_parser.valid_filename(subtitle.release_name):
                subtitle.release_name = string_parser.fix_filename(subtitle.release_name)
            io_file_system.download_subtitle(subtitle, self.download_number, self.download_index_size)
            io_file_system.extract_files_in_dir(VIDEO_FILE.tmp_dir, VIDEO_FILE.subs_dir)
            self.update_text(selection, "✓", subtitle, cfg.color.green)
            self.download_number += 1
            self.download_index_size += 1
            self.downloaded_subtitle.append(subtitle)
        except Exception as e:
            io_log.stdout(str(e), level="error")
            self.update_text(selection, "⨯", subtitle, cfg.color.red)
            self.failed_subtitle_downloads.append(subtitle)
        finally:
            self.sub_listbox.bind("<ButtonPress-1>", self.mouse_b1_press)

    def update_text(self, selection: int, symbol: str, subtitle: Subtitle, color: str) -> None:
        self.sub_listbox.delete(int(selection))
        self.sub_listbox.insert(int(selection), f"{symbol} {subtitle.release_name}")
        self.sub_listbox.itemconfig(int(selection), {"fg": color})



import dataclasses
import logging
from pathlib import Path
from typing import Any, Optional, TypeVar

from subsearch.data.constants import FILE_PATHS
from subsearch.utils import decorators

DATACLASS = TypeVar("DATACLASS")


@decorators.singleton
class Logger:
    def __init__(self, *args, **kwargs) -> None:
        debug_log_file = FILE_PATHS.log
        self.debug_logger = self.create_logger(debug_log_file)

    def create_logger(self, log_file: Path) -> logging.Logger:
        logger = logging.getLogger("subsearch")
        logger.setLevel(logging.DEBUG)
        file_handler = logging.FileHandler(log_file, mode="w")
        file_handler.setLevel(logging.DEBUG)
        formatter = logging.Formatter("%(asctime)s - %(levelname)s - %(message)s", datefmt="%d-%b-%y %H:%M:%S")
        file_handler.setFormatter(formatter)
        logger.addHandler(file_handler)
        return logger

    def close_log_file(self) -> None:
        for handler in self.debug_logger.handlers:
            if isinstance(handler, logging.FileHandler):
                handler.close()

    @decorators.thread_safe_log
    def log(self, message: str, level: str, print_allowed: bool = True) -> None:
        log_methods = {
            "debug": self.debug_logger.debug,
            "info": self.debug_logger.info,
            "warning": self.debug_logger.warning,
            "error": self.debug_logger.error,
            "critical": self.debug_logger.critical,
        }

        log_methods[level](message)
        if print_allowed:
            print(message)


def stdout(message: str | Any, level: str = "info", **kwargs) -> None:
    print_allowed = kwargs.get("print_allowed", True)
    end_new_line = kwargs.get("end_new_line", False)
    _logger = Logger()
    _logger.log(message, level, print_allowed)
    if end_new_line:
        _logger.log("", level, print_allowed)


def stdout_in_brackets(message: str, **kwargs) -> None:
    stdout(f"--- [{message}] ---", **kwargs)


def stdout_match(provider: str, subtitle_name: str, result: int, threshold: int, **kwargs) -> None:
    if result >= threshold:
        stdout(f"> {provider:<14}{result:>3}% {subtitle_name}", kwargs.get("level", "info"))
    else:
        stdout(f"  {provider:<14}{result:>3}% {subtitle_name}", kwargs.get("level", "info"))


def stdout_path_action(action_type: str, src: Path, dst: Optional[Path] = None, **kwargs) -> None:
    if src.is_file():
        type_ = "file"
    elif src.is_dir():
        type_ = "directory"
    else:
        return None

    __src = src.relative_to(src.parent.parent) if src else None
    __dst = dst.relative_to(dst.parent.parent) if dst else None

    action_messages = {
        "remove": rf"Removing {type_}: ...\{__src}",
        "rename": rf"Renaming {type_}: ...\{__src} -> ...\{__dst}",
        "move": rf"Moving {type_}: ...\{__src} -> ...\{__dst}",
        "extract": rf"Extracting archive: ...\{__src} -> ...\{__dst}",
    }

    message = action_messages.get(action_type)

    if not message:
        raise ValueError("Invalid action type")

    stdout(message, **kwargs)


def stdout_dataclass(instance: DATACLASS, **kwargs) -> None:
    if not dataclasses.is_dataclass(instance):
        raise ValueError("Input is not a dataclass instance.")
    stdout_in_brackets(instance.__class__.__name__, **kwargs)
    for field in dataclasses.fields(instance):
        key = field.name
        value = getattr(instance, key)
        padding = " " * (30 - len(key))
        stdout(f"{key}:{padding}{value}", **kwargs)



import re

import cloudscraper
from packaging.version import Version

from subsearch.data import __guid__, __version__


def find_semantic_version(version: str) -> str:
    # semantic expression https://regex101.com/r/M4qItH/
    pattern = r'(?<=__version__ = ")(\d+\.\d+\.\d+[a-zA-Z]*\d*).*?(?=")'
    version_semantic = "".join(re.findall(pattern, version)[0])
    return version_semantic


def scrape_github() -> str:
    scraper = cloudscraper.create_scraper(browser={"browser": "chrome", "platform": "android", "desktop": False})
    url = "https://raw.githubusercontent.com/vagabondHustler/subsearch/main/src/subsearch/data/version.py"
    source = scraper.get(url)
    scontent = source.content
    file_content = str(scontent)
    return file_content


def get_latest_version() -> str:
    version_github = scrape_github()
    return find_semantic_version(version_github)


def is_new_version_avail() -> tuple[bool, bool]:
    new_repo_avail, repo_is_prerelease = False, False
    released_version = get_latest_version()

    if Version(__version__) < Version(released_version):
        if Version(released_version).is_prerelease:
            repo_is_prerelease = True
        new_repo_avail = True

    return new_repo_avail, repo_is_prerelease


# Work in progress
# def find_sha256(url):
#     response = requests.get(url)
#     html_content = response.text

#     parser = selectolax.parser.HTMLParser(html_content)
#     selector_lst = [
#         ".clearfix",
#         "div:nth-child(3)",
#         "section:nth-child(1)",
#         "div:nth-child(2)",
#         "div:nth-child(2)",
#         "div:nth-child(1)",
#         "div:nth-child(1)",
#         "div:nth-child(2)",
#         "h6:nth-child(5)",
#         "p:nth-child(2)",
#     ]

#     css_selector = " > ".join(selector_lst)

#     element = parser.css(css_selector)
#     match = re.search(r"([A-Fa-f0-9]{64})", element[0].html)
#     return match[0]



from selectolax.parser import Node

from subsearch.data.data_classes import Subtitle
from subsearch.providers import core_provider


class SubsceneScraper(core_provider.ProviderHelper):
    def __init__(self, **kwargs) -> None:
        core_provider.ProviderHelper.__init__(self, **kwargs)

    def find_title(self, url: str, current_language: str, definitive_match: list[str]) -> str | None:
        tree = core_provider.get_html_parser(url)
        products = tree.css("div.title")
        for item in products:
            node = item.css_first("a")
            result_href = node.attributes["href"]
            result_title = str(node.child.html).lower()
            if result_title not in definitive_match:
                continue
            current_language = current_language.lower()
            return f"https://subscene.com{result_href}/{current_language}"

    def skip_item(self, item: Node, hi_sub: bool, regular_sub: bool) -> bool:
        if item.css_matches("td.banner-inlist"):
            return True
        if (hi_sub and regular_sub) or (hi_sub is False and regular_sub is False):
            pass
        elif hi_sub is False and item.css_matches("td.a40"):
            return True
        elif regular_sub is False and item.css_matches("td.a41"):
            return True
        return False

    def find_subtitles(self, url: str, hi_sub: bool, regular_sub: bool, subscene_header) -> dict[str, str]:
        subtitles: dict[str, str] = {}
        tree = core_provider.get_html_parser(url, subscene_header)
        products = tree.css("tr")
        for item in products[1:]:
            if self.skip_item(item, hi_sub, regular_sub):
                continue
            node = item.css_first("a")
            if node.child.text_content == "upload":
                continue
            subtitle_href = node.attributes["href"]
            filename = item.css_first("span:nth-child(2)").child.text_content.strip()
            subtitles[filename] = f"https://subscene.com{subtitle_href}"
        return subtitles


class Subscene(SubsceneScraper):
    def __init__(self, **kwargs):
        SubsceneScraper.__init__(self, **kwargs)
        self.provider_name = self.__class__.__name__.lower()

    def start_search(self):
        custom_subscene_header = core_provider.CustomSubsceneHeader(self.app_config)
        header = custom_subscene_header.create_header()
        definitive_match = self._definitive_match()
        found_title_url = self.find_title(self.url_subscene, self.current_language, definitive_match)

        if not found_title_url:
            return []

        subtitle_data = self.find_subtitles(found_title_url, self.hi_sub, self.non_hi_sub, header)
        self._process_subtitle_data(self.provider_name, subtitle_data)

    def _definitive_match(self) -> list[str]:
        if self.tvseries:
            return [f"{self.title} - {self.season_ordinal} season"]
        return [f"{self.title} ({self.year})", f"{self.title} ({(self.year-1)})"]

    @property
    def accepted_subtitles(self) -> list[Subtitle]:
        return self._accepted_subtitles

    @property
    def rejected_subtitles(self) -> list[Subtitle]:
        return self._rejected_subtitles



import ctypes
import time
from pathlib import Path

from subsearch.data.constants import APP_PATHS, DEVICE_INFO, FILE_PATHS, VIDEO_FILE
from subsearch.data.data_classes import Subtitle
from subsearch.gui import screen_manager, system_tray
from subsearch.gui.screens import download_manager
from subsearch.providers import opensubtitles, subscene, yifysubtitles
from subsearch.utils import (
    decorators,
    io_file_system,
    io_log,
    io_toml,
    state_manager,
    string_parser,
)


class Initializer:
    def __init__(self, pref_counter: float) -> None:
        self.file_exist = True if VIDEO_FILE else False
        self.setup_file_system()
        state_manager.CoreStateManager()
        self.core_state = state_manager.CoreStateManager()
        self.start = pref_counter
        self.core_state.set_state(self.core_state.state.INITIALIZE)

        self.app_config = io_toml.get_app_config(FILE_PATHS.config)
        io_log.stdout_dataclass(DEVICE_INFO, level="debug", print_allowed=False)
        io_log.stdout_dataclass(self.app_config, level="debug", print_allowed=False)
        decorators.enable_system_tray = self.app_config.system_tray
        self.system_tray = system_tray.SystemTray()
        self.system_tray.start()

        if self.file_exist:
            io_log.stdout_dataclass(VIDEO_FILE, level="debug", print_allowed=False)
            io_file_system.create_directory(VIDEO_FILE.file_directory)

        self.subtitles_found = 0
        self.ran_download_tab = False
        self.accepted_subtitles: list[Subtitle] = []
        self.rejected_subtitles: list[Subtitle] = []
        self.manually_accepted_subtitles: list[Subtitle] = []
        self.language_data = io_toml.load_toml_data(FILE_PATHS.language_data)

        if self.file_exist:
            self.release_data = string_parser.get_release_data(VIDEO_FILE.filename)
            io_log.stdout_dataclass(self.release_data, level="debug", print_allowed=False)
            provider_urls = string_parser.CreateProviderUrls(self.app_config, self.release_data, self.language_data)
            self.provider_urls = provider_urls.retrieve_urls()
            io_log.stdout_dataclass(self.provider_urls, level="debug", print_allowed=False)
            self.search_kwargs = dict(
                release_data=self.release_data,
                app_config=self.app_config,
                provider_urls=self.provider_urls,
                language_data=self.language_data,
            )

        self.core_state.set_state(self.core_state.state.INITIALIZED)
        if not self.file_exist:
            self.core_state.set_state(self.core_state.state.NO_FILE)

    def setup_file_system(self) -> None:
        io_file_system.create_directory(APP_PATHS.tmp_dir)
        io_file_system.create_directory(APP_PATHS.appdata_subsearch)
        io_toml.resolve_on_integrity_failure()
        if self.file_exist:
            io_file_system.create_directory(VIDEO_FILE.subs_dir)
            io_file_system.create_directory(VIDEO_FILE.tmp_dir)

    def all_providers_disabled(self) -> bool:
        if (
            self.app_config.providers["subscene_site"] is False
            and self.app_config.providers["opensubtitles_site"] is False
            and self.app_config.providers["opensubtitles_hash"] is False
            and self.app_config.providers["yifysubtitles_site"] is False
        ):
            return True
        return False


class SubsearchCore(Initializer):
    def __init__(self, pref_counter: float) -> None:
        Initializer.__init__(self, pref_counter)
        ctypes.windll.kernel32.SetConsoleTitleW(f"subsearch - {DEVICE_INFO.subsearch}")
        if not self.file_exist:
            self.core_state.set_state(self.core_state.state.GUI)
            screen_manager.open_screen("search_options")
            self.core_state.set_state(self.core_state.state.EXIT)
            return None

        if " " in VIDEO_FILE.filename:
            io_log.stdout(f"{VIDEO_FILE.filename} contains spaces, result may vary", level="warning")

        if not self.all_providers_disabled():
            io_log.stdout_in_brackets("Search started")

    def _search_subtitles(self, **kwargs) -> None:
        flag = kwargs.get("flag")
        self.core_state.set_state(kwargs["state_obj"]["call_state"])
        provider_state = kwargs["state_obj"]["state_manager"]()
        provider_state.set_state(provider_state.state.WORKING)
        search_provider = kwargs["provider"](**self.search_kwargs)
        if flag:
            search_provider.start_search(flag=flag)
        else:
            search_provider.start_search()
        self.accepted_subtitles.extend(search_provider.accepted_subtitles)
        self.rejected_subtitles.extend(search_provider.rejected_subtitles)
        provider_state.set_state(provider_state.state.FINNISHED)

    @decorators.call_conditions
    def opensubtitles(self) -> None:
        state_obj = {
            "call_state": self.core_state.state.CALL_OPENSUBTITLES,
            "state_manager": state_manager.OpenSubtitlesStateManager,
        }
        self._search_subtitles(state_obj=state_obj, provider=opensubtitles.OpenSubtitles, flag="hash")
        self._search_subtitles(state_obj=state_obj, provider=opensubtitles.OpenSubtitles, flag="site")

    @decorators.call_conditions
    def subscene(self) -> None:
        state_obj = {
            "call_state": self.core_state.state.CALL_SUBSCENE,
            "state_manager": state_manager.SubsceneStateManager,
        }
        self._search_subtitles(state_obj=state_obj, provider=subscene.Subscene)

    @decorators.call_conditions
    def yifysubtitles(self) -> None:
        state_obj = {
            "call_state": self.core_state.state.CALL_YIFYSUBTITLES,
            "state_manager": state_manager.YifySubtitlesStateManager,
        }
        self._search_subtitles(state_obj=state_obj, provider=yifysubtitles.YifiSubtitles)

    @decorators.call_conditions
    def download_files(self) -> None:
        io_log.stdout_in_brackets(f"Downloading subtitles")
        self.core_state.set_state(self.core_state.state.DOWNLOAD_FILES)
        index_size = len(self.accepted_subtitles)
        for enum, subtitle in enumerate(self.accepted_subtitles, 1):
            io_file_system.download_subtitle(subtitle, enum, index_size)
        self.subtitles_found = index_size
        io_log.stdout("Done with task", level="info", end_new_line=True)

    @decorators.call_conditions
    def manual_download(self) -> None:
        io_log.stdout_in_brackets(f"Manual download")
        self.core_state.set_state(self.core_state.state.MANUAL_DOWNLOAD)
        screen_manager.open_screen("download_manager", subtitles=self.rejected_subtitles)
        self.manually_accepted_subtitles.extend(download_manager.DownloadManager.downloaded_subtitle)
        self.subtitles_found += len(self.manually_accepted_subtitles)
        io_log.stdout("Done with task", level="info", end_new_line=True)

    @decorators.call_conditions
    def extract_files(self) -> None:
        io_log.stdout_in_brackets("Extracting downloads")
        self.core_state.set_state(self.core_state.state.EXTRACT_FILES)
        io_file_system.extract_files_in_dir(VIDEO_FILE.tmp_dir, VIDEO_FILE.subs_dir)
        io_log.stdout("Done with task", level="info", end_new_line=True)

    @decorators.call_conditions
    def subtitle_post_processing(self):
        target = self.app_config.subtitle_post_processing["target_path"]
        resolution = self.app_config.subtitle_post_processing["path_resolution"]
        target_path = io_file_system.create_path_from_string(target, resolution)
        self.subtitle_rename()
        self.subtitle_move_best(target_path)
        self.subtitle_move_all(target_path)

    @decorators.call_conditions
    def subtitle_rename(self) -> None:
        io_log.stdout_in_brackets("Renaming best match")
        self.core_state.set_state(self.core_state.state.AUTOLOAD_RENAME)
        new_name = io_file_system.autoload_rename(VIDEO_FILE.filename, ".srt")
        self.autoload_src = new_name
        io_log.stdout("Done with task", level="info", end_new_line=True)

    @decorators.call_conditions
    def subtitle_move_best(self, target: Path) -> None:
        io_log.stdout_in_brackets("Move best match")
        self.core_state.set_state(self.core_state.state.AUTOLOAD_MOVE)

        io_file_system.move_and_replace(self.autoload_src, target)
        io_log.stdout("Done with task", level="info", end_new_line=True)

    @decorators.call_conditions
    def subtitle_move_all(self, target: Path) -> None:
        io_log.stdout_in_brackets("Move all")
        self.core_state.set_state(self.core_state.state.AUTOLOAD_MOVE_ALL)
        io_file_system.move_all(VIDEO_FILE.subs_dir, target)
        io_log.stdout("Done with task", level="info", end_new_line=True)

    @decorators.call_conditions
    def summary_notification(self, elapsed) -> None:
        io_log.stdout_in_brackets("Summary toast")
        self.core_state.set_state(self.core_state.state.SUMMARY_TOAST)
        elapsed_summary = f"Finished in {elapsed} seconds"
        tot_num_of_subtitles = len(self.accepted_subtitles) + len(self.rejected_subtitles)
        download_summary = f"Matches found {self.subtitles_found}/{tot_num_of_subtitles}"
        if self.subtitles_found > 0:
            msg = "Search Succeeded", f"{download_summary}\n{elapsed_summary}"
            self.system_tray.display_toast(*msg)
        elif self.subtitles_found == 0:
            msg = "Search Failed", f"{download_summary}\n{elapsed_summary}"
            self.system_tray.display_toast(*msg)

    @decorators.call_conditions
    def clean_up(self) -> None:
        io_log.stdout_in_brackets("Cleaning up")
        self.core_state.set_state(self.core_state.state.CLEAN_UP)
        io_file_system.del_file_type(VIDEO_FILE.subs_dir, ".nfo")
        io_file_system.del_directory_content(APP_PATHS.tmp_dir)
        io_file_system.del_directory(VIDEO_FILE.tmp_dir)
        if io_file_system.directory_is_empty(VIDEO_FILE.file_directory):
            io_file_system.del_directory(VIDEO_FILE.subs_dir)
        io_log.stdout("Done with task", level="info", end_new_line=True)

    def core_on_exit(self) -> None:
        elapsed = time.perf_counter() - self.start
        self.summary_notification(elapsed)
        io_log.stdout(f"Finished in {elapsed} seconds")
        self.system_tray.stop()
        if not self.app_config.show_terminal:
            return None
        if DEVICE_INFO.mode == "executable":
            return None

        try:
            input("Enter to exit")
        except KeyboardInterrupt:
            pass
